/* inffast.c -- fast decoding
 * Copyright (C) 1995-2003 Mark Adler
 * For conditions of distribution and use, see copyright notice in zlib.h
 */

#include "zlib.h"


/*
   Decode literal, length, and distance codes and write pOut the resulting
   literal and match bytes until either not enough input or output is
   available, an end-of-block is encountered, or a data error is encountered.
   When large enough input and output buffers are supplied to inflate(), for
   example, a 16K input buffer and a 64K output buffer, more than 95% of the
   inflate execution time is spent in c routine.

   Entry assumptions:

        strm->state.mode == LEN
        strm->avail_in >= 6
        strm->avail_out >= 258
        start >= strm->avail_out
        strm->state.bits < 8

   On return, strm->state.mode is one of:

        LEN -- ran pOut of enough output space or enough available input
        TYPE -- reached end of block code, inflate() to interpret next block
        BAD -- error in block data

   Notes:

    - The maximum input bits used by a length/distance pair is 15 bits for the
      length code, 5 bits for the length extra, 15 bits for the distance code,
      and 13 bits for the distance extra.  This totals 48 bits, or six bytes.
      Therefore if strm->avail_in >= 6, then there is enough input to avoid
      checking for available input while decoding.

    - The maximum bytes that a single length/distance pair can output is 258
      bytes, which is the maximum length that can be coded.  inflate_fast()
      requires strm->avail_out >= 258 for each loop to avoid checking for
      output space.
 */

int inflate_fast(lpzstream strm)
{
	// ref variables to free up a register in the 'do' loop
	const code* lencode = strm->state.lencode;
	const code* distcode = strm->state.distcode;

	// mask for first level of length & distance codes
	const int lmask = ((1U << strm->state.lenbits) - 1);
	const int dmask = ((1U << strm->state.distbits) - 1);
	
	// while in < last, enough input available
	const unsigned char* last = strm->next_in + (strm->avail_in - 5);
	const unsigned char* end = strm->next_out + strm->avail_out - 257;


	// local strm->next_in and strm->next_out
	unsigned char* in  = strm->next_in;
	unsigned char* pOut = strm->next_out;
	
	// local strm->hold and strm->bits
	unsigned long hold = strm->state.hold;
	int bits = strm->state.bits;

	code		c;		// retrieved table entry
	int		len;
	unsigned char*	pSrc;

	// decode literals and length/distances until end-of-block or not enough
	// input data or output space
	int nResult = 0;
	do
	{
		if (bits < 16)
		{
			unsigned long x = *((unsigned long*)in);
			x &= 0xffff;
			x <<= bits;
			hold += x;
			bits += 16;
			in += 2;
		}
		
		c = lencode[hold & lmask];
      
dolen:
		hold >>= c.bits;
		bits -= c.bits;
		assert(bits >= 0);

		if (c.op == 0)  // literal
		{
			*pOut++ = static_cast<unsigned char>(c.val);
		}
		else if (c.op & 16)  // length base 
		{
			len = c.val;
			c.op &= 15;	// number of extra bits
			
			if (c.op)
			{
				if (bits < c.op)
				{
					unsigned long x = *((unsigned long*)in);
					x &= 0xffff;
					x <<= bits;
					hold += x;
					bits += 16;
					in += 2;
				}

				len += (hold & ((1U << c.op) - 1));
				hold >>= c.op;
				bits -= c.op;
			}

			if (bits < 16)
			{
				unsigned long x = *((unsigned long*)in);
				x &= 0xffff;
				x <<= bits;
				hold += x;
				bits += 16;
				in += 2;
			}

			c = distcode[hold & dmask];

dodist:
			hold >>= c.bits;
			bits -= c.bits;
			assert(bits >= 0);
				
			if (c.op & 16)  // distance base
			{
				c.op &= 15;	// number of extra bits

				int dist = c.val;
				if ((unsigned char)bits < c.op)
				{
					unsigned long x = *((unsigned long*)in);
					x &= 0xffff;
					x <<= bits;
					hold += x;
					bits += 16;
					in += 2;
				}

				dist += (hold & ((1U << c.op) - 1));
				hold >>= c.op;
				bits -= c.op;

				// copy directly from output
				pSrc = (pOut - dist);
				if ((pOut - pSrc) <= len || len < 16)
				{
					__asm
					{
						mov	edi, dword ptr pOut;
						mov	esi, dword ptr pSrc;
						mov	ecx, dword ptr len;
						rep	movsb;
					}
				}
				else
				{
					__asm
					{
						mov	ecx, dword ptr len;
						mov	edi, dword ptr pOut;
						mov	esi, dword ptr pSrc;
						mov	eax, ecx;
						shr	ecx, 2;
						rep	movsd;
						mov	ecx, eax;
						and	ecx, 3;
						rep	movsb;
					}
				}
				pOut += len;
			}
			else if ((c.op & 64) == 0)	// 2nd level distance code
			{
				c = distcode[c.val + (hold & ((1U << c.op) - 1))];
				goto dodist;
			}
		}
		else if ((c.op & 64) == 0)	// 2nd level length code
		{
			c = lencode[c.val + (hold & ((1U << c.op) - 1))];
			goto dolen;
		}
		else if (c.op & 32)
		{
			nResult = TYPE;
			break;
		}
	} while(in < last);

	// return unused bytes (on entry, bits < 8, so in won't go too far back)
	len = (bits >> 3);
	in -= len;
	bits -= (len << 3);
	assert(bits >= 0);
	hold &= ((1U << bits) - 1);

	// update state and return
	strm->next_in = in;
	strm->avail_in = 5 - (in - last);
	strm->avail_out = ((strm->next_out + strm->avail_out) - pOut);
	strm->next_out = pOut;
	strm->state.hold = hold;
	strm->state.bits = bits;

	return nResult;
}
