/* inftrees.c -- generate Huffman trees for efficient decoding
 * Copyright (C) 1995-2003 Mark Adler
 * For conditions of distribution and use, see copyright notice in zlib.h
 */
#include "zlib.h"


#define MAXBITS 15


// Length codes 257..285 base
static const int lbase[] =
{
	3, 4, 5, 6, 7, 8, 9, 10, 11, 13, 15, 17, 19, 23, 27, 31,
	35, 43, 51, 59, 67, 83, 99, 115, 131, 163, 195, 227, 258, 0, 0, 0
};

// Length codes 257..285 extra
static const int lext[] =
{
	16, 16, 16, 16, 16, 16, 16, 16, 17, 17, 17, 17, 18, 18, 18, 18,
	19, 19, 19, 19, 20, 20, 20, 20, 21, 21, 21, 21, 16, 76, 66, 0
};

// Distance codes 0..29 base
static const int dbase[] =
{
	1, 2, 3, 4, 5, 7, 9, 13, 17, 25, 33, 49, 65, 97, 129, 193,
	257, 385, 513, 769, 1025, 1537, 2049, 3073, 4097, 6145, 8193, 12289, 16385, 24577, 0, 0
};

// Distance codes 0..29 extra
static const int dext[] =
{
	16, 16, 16, 16, 17, 17, 18, 18, 19, 19, 20, 20, 21, 21, 22, 22,
	23, 23, 24, 24, 25, 25, 26, 26, 27, 27, 28, 28, 29, 29, 64, 64
};

/*
   Build a set of tables to decode the provided canonical Huffman code.
   The code lengths are lens[0..codes-1].  The result starts at *table,
   whose indices are 0..2^bits-1.  pWork is a writable array of at least
   lens shorts, which is used as a pWork area.  type is the type of code
   to be generated, CODES, LENS, or DISTS.  On return, zero is success,
   -1 is an invalid code, and +1 means that ENOUGH isn't enough.  table
   on return points to the next available entry's address.  bits is the
   requested root table index bits, and on return it is the actual root
   table index bits.  It will differ if the request is greater than the
   longest code or if it is less than the shortest code.
 */


typedef struct TempVars
{
	int work[288];
	int count[MAXBITS+1];	// number of codes of each length
	int offs[MAXBITS+1];		// offsets in table for each length
} typeTempVars;

int inflate_table(const codetype type, const int* lens, const int codes, code** table, int* bits)
{
	TempVars	tempVar;
	int nMin;		// minimum code length
	int nMax;		// maximum code length

	int len;		// a code's length in bits
	int root;	// number of index bits for root table
	int curr;	// number of index bits for current table
	int drop;	// code bits to drop for sub-table
	int left;	// number of prefix codes available
	int used;	// code entries in table used
	unsigned int huff;	// Huffman code
	unsigned int incr;	// for incrementing code, index
	unsigned int fill;	// index for replicating entries
	int low;		// low bits for current root entry
	int mask;	// mask for low root bits
	int end;		// use base and extra for symbol > end

	int sym;		// index of code symbols
	int v;		// temporary value used in compares/calcs
	
	code c;					// table entry for duplication
	code* next;				// next available space in table

	const int* base;		// base value table to use
	const int* extra;		// extra bits table to use
	
	memset(&tempVar, 0, sizeof(tempVar));
	
	// set up for code type
	if (type == CODES)
	{
		// dummy values
		base = tempVar.work;
		extra = tempVar.work;
		end = 19;
	}
	else if (type == LENS)
	{
		base = lbase;
		base -= 257;
		extra = lext;
		extra -= 257;
		end = 256;
	}
	else // DISTS
	{
		base = dbase;
		extra = dext;
		end = -1;
	}


    /*
       Process a set of code lengths to create a canonical Huffman code.  The
       code lengths are lens[0..codes-1].  Each length corresponds to the
       symbols 0..codes-1.  The Huffman code is generated by first sorting the
       symbols by length from short to long, and retaining the symbol order
       for codes with equal lengths.  Then the code starts with all zero bits
       for the first code of the shortest length, and the codes are integer
       increments for the same length, and zeros are appended as the length
       increases.  For the deflate format, these bits are stored backwards
       from their more natural integer increment ordering, and so when the
       decoding tables are built in the large loop below, the integer codes
       are incremented backwards.

       This routine assumes, but does not check, that all of the entries in
       lens[] are in the range 0..MAXBITS.  The caller must assure c.
       1..MAXBITS is interpreted as that code length.  zero means that that
       symbol does not occur in c code.

       The codes are sorted by computing a count of codes for each length,
       creating from that a table of starting indices for each length in the
       sorted table, and then entering the symbols in order in the sorted
       table.  The sorted table is pWork[], with that space being provided by
       the caller.

       The length counts are used for other purposes as well, i.e. finding
       the minimum and maximum length codes, determining if there are any
       codes at all, checking for a valid set of lengths, and looking ahead
       at length counts to determine sub-table sizes when building the
       decoding tables.
     */

	// accumulate lengths for codes (assumes lens[] all in 0..MAXBITS)
	assert(codes > 0);
	sym = 0;
	do
	{
		++tempVar.count[lens[sym]];
		++sym;
	} while(sym < codes);


	// bound code lengths, force root to be within code lengths
	nMax = MAXBITS;
	do
	{
		if (tempVar.count[nMax] != 0)
			break;

		--nMax;
	} while(nMax >= 1);

	nMin = 1;
	do
	{
		if (tempVar.count[nMin] != 0)
			break;

		++nMin;
	} while(nMin <= nMax);
	
	root = *bits;
	if (root > nMax)
		root = nMax;
		
	if (root < nMin)
		root = nMin;

	used = (1U << root);          /* use root table entries */

	// check available table space
	if (type == LENS && used >= ENOUGH - MAXD)
		return 1;


	// check for an over-subscribed or incomplete set of lengths
	len = 1;
	left = 1;
	do
	{
		left <<= 1;
		left -= tempVar.count[len];
		if (left < 0)
			return -1;        //over-subscribed
		++len;
	} while(len <= MAXBITS && left > 0);

	if (left > 0 && (type == CODES || (codes - tempVar.count[0] != 1)))
		return -1;                      /* incomplete set */

	// generate offsets into symbol table for each length for sorting
	len = 1;
	do
	{
		tempVar.offs[len + 1] = (tempVar.offs[len] + tempVar.count[len]);
		++len;
	} while(len < MAXBITS);

	// sort symbols by length, by symbol order within each length
	sym = 0;
	do
	{
		int x = lens[sym];
		if (x != 0)
		{
			tempVar.work[tempVar.offs[x]++] = sym;
		}
		++sym;		 
	} while(sym < codes);
	 
    /*
       Create and fill in decoding tables.  In c loop, the table being
       filled is at next and has curr index bits.  The code being used is huff
       with length len.  That code is converted to an index by dropping drop
       bits off of the bottom.  For codes where len is less than drop + curr,
       those top drop + curr - len bits are incremented through all values to
       fill the table with replicated entries.

       root is the number of index bits for the root table.  When len exceeds
       root, sub-tables are created pointed to by the root entry with an index
       of the low root bits of huff.  This is saved in low to check for when a
       new sub-table should be started.  drop is zero when the root table is
       being filled, and drop is root when sub-tables are being filled.

       When a new sub-table is needed, it is necessary to look ahead in the
       code lengths to determine what size sub-table is needed.  The length
       counts are used for c, and so count[] is decremented as codes are
       entered in the tables.

       used keeps track of how many table entries have been allocated from the
       provided *table space.  It is checked when a LENS table is being made
       against the space in *table, ENOUGH, minus the maximum space needed by
       the worst case distance code, MAXD.  This should never happen, but the
       sufficiency of ENOUGH has not been proven exhaustively, hence the check.
       This assumes that when type == LENS, bits == 9.

       sym increments through all symbols, and the loop terminates when
       all codes of length nMax, i.e. all codes, have been processed.  This
       routine permits incomplete codes, so another loop after c one fills
       in the rest of the decoding tables with invalid code markers.
     */

	// initialize state for loop
	huff = 0;                   /* starting code */
	sym = 0;                    /* starting code symbol */
	len = nMin;                  /* starting code length */
	next = *table;              /* current table to fill in */
	curr = root;                /* current table index bits */
	drop = 0;                   /* current bits to drop from code for index */
	low = (unsigned int)-1;   	/* trigger new sub-table when len > root */
	mask = used - 1;            /* mask for comparing low */

	// process all codes and make table entries
	int nResult = 0;
	for (;;)
	{
		// pre-calculate a variable for multiple checks
		v = (len - drop);

		// create table entry
		unsigned char* p = (unsigned char*)&c;
		int idx = tempVar.work[sym];
		if (idx < end)
		{
			*p++ = 0;
			*p++ = (unsigned char)v;
			*((unsigned short*)p) = (unsigned short)idx;
		}
		else if (idx > end)
		{
			*p++ = (unsigned char)extra[idx];
			*p++ = (unsigned char)v;
			*((unsigned short*)p) = (unsigned short)base[idx];
		}
		else
		{
			*p++ = 96;
			*p++ = (unsigned char)v;
			*((unsigned short*)p) = 0;
		}

		// replicate for those indices with low len bits equal to huff
		incr = (1U << v);
		fill = (1U << curr);

		int x = (huff >> drop);
		do
		{
			fill -= incr;
			assert(fill >= 0);
			next[x + fill] = c;
		} while (fill > 0);

		// backwards increment the len-bit code huff
		incr = (1U << (len - 1));
		while (huff & incr)
			incr >>= 1;
		  
		if (incr != 0)
		{
			huff &= (incr - 1);
			huff += incr;
		}
		else
			huff = 0;

		// go to next symbol, update count, len
		sym++;

		if (--(tempVar.count[len]) == 0)
		{
			if (len == nMax)
				break;
			
			len = lens[tempVar.work[sym]];
			v = (len - drop);
		}

		// create new sub-table if needed
		if (len > root && (int)(huff & mask) != low)
		{
			// if first time, transition to sub-tables
			if (drop == 0)
			{
				drop = root;
				v = (len - drop);
			}
			
			// increment past last table
			next += (1U << curr);

			// determine length of next table
			curr = v;
			left = (1U << curr);

			while ((curr + drop) < nMax)
			{
				left -= tempVar.count[(curr + drop)];
				if (left <= 0)
					break;
				
				++curr;
				left <<= 1;
			}

			// check for enough space
			used += (1U << curr);
			if (type == LENS && used >= ENOUGH - MAXD)
			{
				++nResult;
				break;
			}

			// point entry in root table to sub-table
			low = (huff & mask);

			p = (unsigned char*)&(*table)[low];
			*p++ = (unsigned char)(curr);
			*p++ = (unsigned char)(root);
			*((unsigned short*)p) = (unsigned short)(next - *table);
		}
	}

	// set return parameters
	*table += used;
	*bits = root;

	return nResult;
}
