#if !defined(__ZLIB_H__)
#define __ZLIB_H__

#if defined(_DEBUG)
#include <assert.h>
#elif !defined(assert)
	#define assert(p)
#endif

#include <memory.h>


#define Z_NO_FLUSH      0
#define Z_PARTIAL_FLUSH 1 /* will be removed, use Z_SYNC_FLUSH instead */
#define Z_SYNC_FLUSH    2
#define Z_FULL_FLUSH    3
#define Z_FINISH        4
#define Z_BLOCK         5
/* Allowed flush values; see deflate() and inflate() below for details */

#define Z_OK            0
#define Z_STREAM_END    1
#define Z_NEED_DICT     2
#define Z_ERRNO        (-1)
#define Z_STREAM_ERROR (-2)
#define Z_DATA_ERROR   (-3)
#define Z_MEM_ERROR    (-4)
#define Z_BUF_ERROR    (-5)
#define Z_VERSION_ERROR (-6)
/* Return codes for the compression/decompression functions. Negative
 * values are errors, positive values are used for special but normal events.
 */

#define Z_NO_COMPRESSION         0
#define Z_BEST_SPEED             1
#define Z_BEST_COMPRESSION       9
#define Z_DEFAULT_COMPRESSION  (-1)
/* compression levels */

#define Z_FILTERED            1
#define Z_HUFFMAN_ONLY        2
#define Z_RLE                 3
#define Z_DEFAULT_STRATEGY    0
/* compression strategy; see deflateInit2() below for details */

#define Z_BINARY   0
#define Z_ASCII    1
#define Z_UNKNOWN  2
/* Possible values of the data_type field (though see inflate()) */

#define Z_DEFLATED   8
/* The deflate compression method (the only one supported in this version) */

#define Z_NULL  0

#define MAX_WBITS		15

/* Possible inflate modes between inflate() calls */
typedef enum eTableCode
{
	HEAD = 0,       /* i: waiting for magic header */
	FLAGS,      /* i: waiting for method and flags (gzip) */
	TIME,       /* i: waiting for modification time (gzip) */
	OS,         /* i: waiting for extra flags and operating system (gzip) */
	EXLEN,      /* i: waiting for extra length (gzip) */
	EXTRA,      /* i: waiting for extra bytes (gzip) */
	NAME,       /* i: waiting for end of file name (gzip) */
	COMMENT,    /* i: waiting for end of comment (gzip) */
	HCRC,       /* i: waiting for header crc (gzip) */
	DICTID,     /* i: waiting for dictionary check value */
	DICT,       /* waiting for inflateSetDictionary() call */
	TYPE,       /* i: waiting for type bits, including last-flag bit */
	TYPEDO,     /* i: same, but skip check to exit inflate on new block */
	STORED,     /* i: waiting for stored size (length and complement) */
	COPY,       /* i/o: waiting for input or output to copy stored block */
	TABLE,      /* i: waiting for dynamic block table lengths */
	LENLENS,    /* i: waiting for code length code lengths */
	CODELENS,   /* i: waiting for length/lit and distance code lengths */
	LEN,        /* i: waiting for length/lit code */
	LENEXT,     /* i: waiting for length extra bits */
	DIST,       /* i: waiting for distance code */
	DISTEXT,    /* i: waiting for distance extra bits */
	MATCH,      /* o: waiting for output space to copy string */
	LIT,        /* o: waiting for output space to write literal */
	CHECK,      /* i: waiting for 32-bit check value */
	LENGTH,     /* i: waiting for 32-bit length (gzip) */
	DONE,       /* finished check, done -- remain here until reset */
	BAD,        /* got a data error -- remain here until reset */
	MEM,        /* got an inflate() memory error -- remain here until reset */
	SYNC        /* looking for synchronization bytes to restart inflate() */
} inflate_mode;

/*
    State transitions between above modes -

    (most modes can go to the BAD or MEM mode -- not shown for clarity)

    Process header:
        HEAD -> (gzip) or (zlib)
        (gzip) -> FLAGS -> TIME -> OS -> EXLEN -> EXTRA -> NAME
        NAME -> COMMENT -> HCRC -> TYPE
        (zlib) -> DICTID or TYPE
        DICTID -> DICT -> TYPE
    Read deflate blocks:
            TYPE -> STORED or TABLE or LEN or CHECK
            STORED -> COPY -> TYPE
            TABLE -> LENLENS -> CODELENS -> LEN
    Read deflate codes:
                LEN -> LENEXT or LIT or TYPE
                LENEXT -> DIST -> DISTEXT -> MATCH -> LEN
                LIT -> LEN
    Process trailer:
        CHECK -> LENGTH -> DONE
 */

/* Structure for decoding tables.  Each entry provides either the
   information needed to do the operation requested by the code that
   indexed that table entry, or it provides a pointer to another
   table that indexes more bits of the code.  op indicates whether
   the entry is a pointer to another table, a literal, a length or
   distance, an end-of-block, or an invalid code.  For a table
   pointer, the low four bits of op is the number of index bits of
   that table.  For a length or distance, the low four bits of op
   is the number of extra bits to get after the code.  bits is
   the number of bits in this code or part of the code to drop off
   of the bit buffer.  val is the actual byte to output in the case
   of a literal, the base length or distance, or the offset from
   the current table to the next table.  Each entry is four bytes. */

typedef struct
{
	unsigned char op;           /* operation, extra bits, table bits */
	unsigned char bits;         /* bits in this part of the code */
	unsigned short val;         /* offset in table or code value */
} code;


/* op values as set by inflate_table():
    00000000 - literal
    0000tttt - table link, tttt != 0 is the number of table index bits
    0001eeee - length or distance, eeee is the number of extra bits
    01100000 - end of block
    01000000 - invalid code
 */

// Maximum size of dynamic tree.  The maximum found in a long but non-
// exhaustive search was 1004 code structures (850 for length/literals
// and 154 for distances, the latter actually the result of an
// exhaustive search).  The true maximum is not known, but the value
// below is more than safe.
#define ENOUGH 1440
#define MAXD 154

// Type of code to build for infate_table()
typedef enum { CODES, LENS, DISTS } codetype;


// state information for inflate, inflate_table and inflate_fast
typedef struct InflateState
{
	// bit accumulator
	int hold;		// input bit accumulator
	int bits;		// number of bits in "in"
	
	// for string and stored block copying
	int length;		// literal or length of data to copy
	int offset;		// distance back to copy string from
	
	// for table and code decoding
	int extra;		// extra bits needed
	
	// fixed and dynamic code tables
	int lenbits;	// index bits for lencode
	int distbits;	// index bits for distcode
	
	// dynamic table building
	int ncode;		// number of code length code lengths
	int nlen;		// number of length code lengths
	int ndist;		// number of distance code lengths
	int have;		// number of code lengths in lens[]
	
	code* next;		// next available space in codes[]

	code const* lencode;		// starting table for length/literal codes
	code const* distcode;	// starting table for distance codes
} typeInflateState, *lpInflateState;


typedef struct z_stream_s
{
	unsigned char*	next_in;		// next input byte
	unsigned char*	next_out;	// next output byte should be put there

	unsigned long	avail_in;	// number of bytes available at next_in
	unsigned long	avail_out;	// remaining free space at next_out

	unsigned long	total_in;	// total number of input bytes read so far
	unsigned long	total_out;	// total number of bytes output so far

	InflateState state;
} z_stream, *lpzstream;

//int inflate(const void* pInBuf, const unsigned long dwBufLen, void* pOutBuf, const unsigned long dwOutLen);

#define __ASM_INFLATE
#if defined(__ASM_INFLATE)
	#if defined(__cplusplus)
	extern "C" {
	#endif
		int __stdcall inflate_fast(lpzstream pStream);
		int __stdcall inflate_table(const codetype type, const int* lens, const int codes, code** table, int* bits);
	#if defined(__cplusplus)
	}
	#endif
#else
	int inflate_fast(lpzstream strm);
	int inflate_table(const codetype type, const int* lens, const int codes, code** table, int* bits);
#endif


#endif	// !defined(__ZLIB_H__)
